from enum import Enum
from typing import Dict

from pydantic import BaseModel


class AwsRegion(str, Enum):
    us_east_1 = "us-east-1"
    us_east_2 = "us-east-2"
    us_west_1 = "us-west-1"
    us_west_2 = "us-west-2"


class AwsCredentialType(str, Enum):
    assumed_role = "assumed_role"
    key_secret = "key_secret"


class AwsCredentials(BaseModel):
    credential_type: AwsCredentialType = AwsCredentialType.key_secret
    region: str = AwsRegion.us_west_2


class BaseAwsIntegrationCreate(AwsCredentials):
    token: Dict[str, str]



from typing import Literal

from pydantic import BaseModel

from galileo_core.schemas.core.integration.integration_type import LLMIntegration


class VertexAIIntegrationCreate(BaseModel):
    name: Literal[LLMIntegration.vertex_ai] = LLMIntegration.vertex_ai
    token: str



from enum import Enum
from typing import Optional, Union

from pydantic import UUID4, BaseModel, Field, ValidationInfo, field_validator

from galileo_core.schemas.protect.ruleset import RulesetsMixin


class StageType(str, Enum):
    local = "local"
    central = "central"


class Stage(BaseModel):
    name: str = Field(description="Name of the stage. Must be unique within the project.")
    project_id: UUID4 = Field(description="ID of the project to which this stage belongs.")
    description: Optional[str] = Field(
        description="Optional human-readable description of the goals of this guardrail.",
        default=None,
    )
    type: StageType = Field(description="Type of the stage.", default=StageType.local)
    paused: bool = Field(
        description="Whether the action is enabled. If False, the action will not be applied.",
        default=False,
    )


class StageWithRulesets(Stage, RulesetsMixin):
    @field_validator("type", mode="before")
    def validate_type(cls, value: Union[StageType, str], info: ValidationInfo) -> StageType:
        stage_type = StageType(value)
        rulesets = info.data.get("prioritized_rulesets") or info.data.get("rulesets") or []
        if value == StageType.local and len(rulesets) > 0:
            raise ValueError("Local stages cannot include ruleset definitions during creation.")
        return stage_type


class CreateStage(Stage):
    created_by: UUID4


class StageDB(CreateStage):
    id: UUID4
    version: Union[int, None] = None



from functools import partial
from typing import List, Optional

from pydantic import UUID4

from galileo_core.constants.request_method import RequestMethod
from galileo_core.constants.routes import Routes
from galileo_core.exceptions.http import GalileoHTTPException
from galileo_core.helpers.logger import logger
from galileo_core.helpers.pagination import paginated_request
from galileo_core.schemas.base_config import GalileoConfig
from galileo_core.schemas.core.collaboration_role import CollaboratorRole
from galileo_core.schemas.core.collaborator import GroupCollaboratorCreate
from galileo_core.schemas.core.integration.group_integration import (
    GroupIntegrationCollaboratorResponse,
)


def list_group_integration_collaborators(
    integration_id: UUID4,
    config: Optional[GalileoConfig] = None,
) -> List[GroupIntegrationCollaboratorResponse]:
    config = config or GalileoConfig.get()
    logger.debug("Listing groups the integration has been shared with…")
    group_integration_collaborators = paginated_request(
        partial(
            config.api_client.request,
            RequestMethod.GET,
            Routes.integration_groups.format(integration_id=integration_id),
        ),
        "collaborators",
    )
    groups = [GroupIntegrationCollaboratorResponse.model_validate(group) for group in group_integration_collaborators]
    logger.debug(f"Listed all groups, found {len(groups)} groups.")
    return groups


def share_integration_with_group(
    integration_id: UUID4,
    group_id: UUID4,
    config: Optional[GalileoConfig] = None,
) -> GroupIntegrationCollaboratorResponse:
    config = config or GalileoConfig.get()
    logger.debug(f"Sharing integration {integration_id} with group {group_id} with role {CollaboratorRole.viewer}...")
    response_dict = config.api_client.request(
        RequestMethod.POST,
        Routes.integration_groups.format(integration_id=integration_id),
        json=[GroupCollaboratorCreate(group_id=group_id).model_dump(mode="json")],
    )
    group_shared = [GroupIntegrationCollaboratorResponse.model_validate(group) for group in response_dict]
    logger.debug(f"Shared integration {integration_id} with group {group_id} with role {CollaboratorRole.viewer}.")

    try:
        return group_shared[0]
    except IndexError:
        raise GalileoHTTPException(
            message="Galileo API returned empty response.",
            status_code=422,
            response_text=f"Response body {group_shared}",
        )


def update_group_integration_collaborator(
    integration_id: UUID4,
    group_id: UUID4,
    role: CollaboratorRole = CollaboratorRole.viewer,
    config: Optional[GalileoConfig] = None,
) -> GroupIntegrationCollaboratorResponse:
    config = config or GalileoConfig.get()
    logger.debug(f"Updating integration {integration_id}: assigning role {role} to group {group_id}…")
    response_dict = config.api_client.request(
        RequestMethod.PATCH,
        Routes.integration_group.format(integration_id=integration_id, group_id=group_id),
        json=GroupCollaboratorCreate(group_id=group_id, role=role).model_dump(mode="json"),
    )
    group_integration = GroupIntegrationCollaboratorResponse.model_validate(response_dict)
    logger.debug(f"Updated integration {integration_id}: assigned role {role} to group {group_id}…")
    return group_integration


def delete_group_integration_collaborator(
    integration_id: UUID4,
    group_id: UUID4,
    config: Optional[GalileoConfig] = None,
) -> None:
    config = config or GalileoConfig.get()
    logger.debug(f"Removing access for group {group_id} from integration {integration_id}")
    config.api_client.request(
        RequestMethod.DELETE, Routes.integration_group.format(integration_id=integration_id, group_id=group_id)
    )
    logger.debug(f"Access for group {group_id} removed from integration {integration_id}.")



from enum import Enum

from typing_extensions import assert_never


class RootType(str, Enum):
    """The root-level type of a logged step hierarchy.

    Maps fine-grained StepType values to the three top-level categories
    used throughout the platform: session, trace, and span.
    """

    session = "session"
    trace = "trace"
    span = "span"

    def pluralize(self) -> str:
        if self == RootType.session:
            return "sessions"
        elif self == RootType.trace:
            return "traces"
        elif self == RootType.span:
            return "spans"
        else:
            assert_never(self)  # pragma: no cover



from json import JSONEncoder
from typing import Any

from pydantic import BaseModel


class PydanticJsonEncoder(JSONEncoder):
    def __init__(
        self,
        *args: Any,
        exclude_none: bool = True,
        exclude_unset: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._exclude_none = exclude_none
        self._exclude_unset = exclude_unset

    def default(self, obj: Any) -> Any:
        if isinstance(obj, BaseModel):
            return obj.model_dump(mode="json", exclude_none=self._exclude_none, exclude_unset=self._exclude_unset)
        return super().default(obj)



from typing import List, Literal, Optional

from pydantic import Field
from pydantic.types import UUID4

from galileo_core.schemas.logging.step import BaseStep, StepType
from galileo_core.schemas.logging.trace import Trace


class BaseSession(BaseStep):
    type: Literal[StepType.session] = Field(
        default=StepType.session, description=BaseStep.model_fields["type"].description
    )
    previous_session_id: Optional[UUID4] = None


class Session(BaseSession):
    traces: List[Trace] = Field(
        default_factory=list,
        description="List of traces associated with this session. Each trace can have its own spans.",
    )


# Define SessionWithTraces as an alias for Session for backward compatibility
SessionWithTraces = Session



from typing import Dict, List, Literal, Optional

from pydantic import BaseModel, ConfigDict, Field

from galileo_core.schemas.core.integration.azure_auth_type import AzureAuthenticationType
from galileo_core.schemas.core.integration.integration_type import LLMIntegration

DEFAULT_AZURE_API_VERSION = "2024-10-21"


class AzureModelDeployment(BaseModel):
    model: str = Field(description="The name of the model.")
    id: str = Field(description="The ID of the deployment.")

    model_config = ConfigDict(protected_namespaces=())


class AzureExtras(BaseModel):
    proxy: bool = False
    endpoint: str
    authentication_type: AzureAuthenticationType = AzureAuthenticationType.api_key
    authentication_scope: Optional[str] = None
    headers: Optional[Dict[str, str]] = None
    api_version: str = DEFAULT_AZURE_API_VERSION
    azure_deployment: Optional[str] = None
    available_deployments: Optional[List[AzureModelDeployment]] = Field(
        default=None,
        description="The available deployments for this integration."
        " If provided, we will not try to get this list from Azure.",
    )


class AzureIntegrationCreate(AzureExtras):
    name: Literal[LLMIntegration.azure] = LLMIntegration.azure
    token: str



from datetime import datetime
from enum import Enum
from typing import Annotated, ClassVar, Generic, Literal, TypeVar

from pydantic import (
    UUID4,
    BaseModel,
    Field,
    ValidationInfo,
    field_validator,
    model_validator,
)


class Operator(str, Enum):
    eq = "eq"
    ne = "ne"
    contains = "contains"
    one_of = "one_of"
    not_in = "not_in"
    gt = "gt"
    gte = "gte"
    lt = "lt"
    lte = "lte"
    between = "between"


class FilterType(str, Enum):
    boolean = "boolean"
    number = "number"
    string = "string"
    enum = "enum"
    collection = "collection"
    map = "map"
    custom_number = "custom_number"
    custom_boolean = "custom_boolean"
    custom_uuid = "custom_uuid"
    id = "id"
    date = "date"
    function = "function"


class FilterBase(BaseModel):
    name: str | None


BooleanOperator = Literal[Operator.eq, Operator.ne]


class BooleanFilter(FilterBase):
    """
    Filters on a boolean field.
    """

    filter_type: ClassVar[Literal[FilterType.boolean]] = FilterType.boolean
    operator: BooleanOperator = Field(default=Operator.eq)
    value: bool


StringOperator = Literal[Operator.eq, Operator.ne, Operator.contains, Operator.one_of, Operator.not_in]


class StringFilter(FilterBase):
    """
    Filters on a string field.
    """

    filter_type: ClassVar[Literal[FilterType.string]] = FilterType.string
    operator: StringOperator
    value: str | list[str]
    case_sensitive: bool = True

    @model_validator(mode="after")
    def validate_operator_value(self) -> "StringFilter":
        operator = self.operator
        value = self.value
        match operator:
            case Operator.one_of | Operator.not_in:
                if not isinstance(value, list):
                    raise ValueError(f"Value must be a list for operator {operator}.")
            case Operator.eq | Operator.ne | Operator.contains:
                if isinstance(value, list):
                    raise ValueError(f"Value must be a string for operator {operator}.")
            case _:
                raise ValueError(f"Invalid operator {operator}.")
        return self


CollectionOperator = Literal[Operator.eq, Operator.contains, Operator.one_of, Operator.not_in]


class CollectionFilter(FilterBase):
    """
    Filters for string items in a collection/list.
    """

    filter_type: ClassVar[Literal[FilterType.collection]] = FilterType.collection
    operator: CollectionOperator
    value: str | list[str]
    case_sensitive: bool = True

    @model_validator(mode="after")
    def validate_operator_value(self) -> "CollectionFilter":
        operator = self.operator
        value = self.value
        match operator:
            case Operator.not_in | Operator.one_of:
                if not isinstance(value, list):
                    raise ValueError(f"Value must be a list for operator {operator}.")
            case Operator.eq | Operator.contains:
                if not isinstance(value, str):
                    raise ValueError(f"Value must be a string for operator {operator}.")
            case _:
                raise ValueError(f"Invalid operator {operator}.")
        return self


MapOperator = Literal[Operator.one_of, Operator.not_in, Operator.eq, Operator.ne]


class MapFilter(FilterBase):
    """
    Filters for string items in a map / dictionary.
    """

    filter_type: ClassVar[Literal[FilterType.map]] = FilterType.map
    operator: MapOperator
    key: str
    value: str | list[str]

    @model_validator(mode="after")
    def validate_operator_value(self) -> "MapFilter":
        operator = self.operator
        value = self.value
        if operator in (Operator.one_of, Operator.not_in):
            if not isinstance(value, list):
                raise ValueError(f"Value must be a list for operator {operator}.")
        elif operator in (Operator.eq, Operator.ne):
            if isinstance(value, list):
                raise ValueError(f"Value must be a string for operator {operator}.")
        return self


EnumOperator = Literal[Operator.eq, Operator.ne, Operator.one_of, Operator.not_in]

E = TypeVar("E", bound="Enum")


class EnumFilter(FilterBase, Generic[E]):
    """
    Filters on a string field, with limited categories.
    """

    filter_type: ClassVar[Literal[FilterType.enum]] = FilterType.enum
    operator: EnumOperator
    value: Annotated[
        E | list[E],
        Field(
            json_schema_extra={
                "anyOf": [
                    {
                        "type": "string",
                        "example": "ENUM_VALUE",
                        "description": "Single enum value - specific options depend on the concrete enum type used",
                    },
                    {
                        "type": "array",
                        "items": {"type": "string", "example": "ENUM_VALUE"},
                        "example": ["ENUM_VALUE_1", "ENUM_VALUE_2"],
                        "description": "Array of enum values",
                    },
                ]
            }
        ),
    ]


IDOperator = Literal[Operator.eq, Operator.ne, Operator.one_of, Operator.not_in, Operator.contains]


class IDFilter(FilterBase):
    """
    Filters on a UUID field.
    """

    filter_type: ClassVar[Literal[FilterType.id]] = FilterType.id
    operator: IDOperator = Field(default=Operator.eq)
    value: UUID4 | list[UUID4 | str] | str

    @field_validator("value", mode="before")
    def validate_value(
        cls, value: UUID4 | list[UUID4 | str] | str, info: ValidationInfo
    ) -> UUID4 | list[UUID4 | str] | str:
        operator = info.data.get("operator", Operator.eq)
        match operator:
            case Operator.one_of | Operator.not_in:
                if not isinstance(value, list):
                    raise ValueError(f"Value must be a list for operator {operator}.")
            case Operator.eq | Operator.ne:
                if isinstance(value, list):
                    raise ValueError(f"Value must be a single UUID for operator {operator}.")
            case Operator.contains:
                if not isinstance(value, str):
                    raise ValueError(f"Value must be a string for operator {operator}.")
            case _:
                raise ValueError(f"Invalid operator {operator}.")
        return value


DateOperator = Literal[Operator.eq, Operator.ne, Operator.gt, Operator.gte, Operator.lt, Operator.lte]


class DateFilter(FilterBase):
    """Filters on a datetime field."""

    filter_type: ClassVar[Literal[FilterType.date]] = FilterType.date
    operator: DateOperator
    value: datetime


NumberOperator = Literal[
    Operator.eq,
    Operator.ne,
    Operator.gt,
    Operator.gte,
    Operator.lt,
    Operator.lte,
    Operator.between,
]


class CustomNumberFilter(FilterBase):
    filter_type: ClassVar[Literal[FilterType.custom_number]] = FilterType.custom_number
    operator: NumberOperator
    value: int | float | list[int] | list[float]


class CustomBooleanFilter(FilterBase):
    filter_type: ClassVar[Literal[FilterType.custom_boolean]] = FilterType.custom_boolean
    value: bool


class CustomFunctionFilter(FilterBase):
    filter_type: ClassVar[Literal[FilterType.function]] = FilterType.function


class CustomUUIDFilter(FilterBase):
    filter_type: ClassVar[Literal[FilterType.custom_uuid]] = FilterType.custom_uuid
    value: UUID4


QueryFilterV2 = Annotated[
    CollectionFilter
    | CustomBooleanFilter
    | CustomNumberFilter
    | CustomUUIDFilter
    | DateFilter
    | EnumFilter
    | IDFilter
    | MapFilter
    | StringFilter
    | BooleanFilter
    | CustomFunctionFilter,
    Field(discriminator="filter_type"),
]



from enum import Enum


class ContentModality(str, Enum):
    """Classification of content modality"""

    text = "text"  # plain text
    document = "document"
    image = "image"
    audio = "audio"
    video = "video"


class FileSource(str, Enum):
    """Source of the file data."""

    direct_upload = "direct_upload"  # Base64 embedded in trace
    external_files_api = "external_files_api"  # Provider file API (OpenAI, Gemini, Anthropic, etc.)
    external_url = "external_url"  # Public URL reference
    assembled_stream = "assembled_stream"  # Streaming audio/video assembled


class ExternalProvider(str, Enum):
    """External provider of file"""

    openai_files_api = "openai_files_api"


class FileStatus(str, Enum):
    """Processing status of the file."""

    complete = "complete"  # Successfully stored
    failed = "failed"  # Storage failed
    pending = "pending"  # Data is still being uploaded
    not_uploaded = "not_uploaded"  # external_url / provider file IDs not uploaded


class MultimodalCapability(str, Enum):
    vision = "vision"
    audio = "audio"


MultimodalCapabilityMapping: dict[MultimodalCapability, set[ContentModality]] = {
    MultimodalCapability.vision: {ContentModality.image, ContentModality.document},
    MultimodalCapability.audio: {ContentModality.audio},
}



from datetime import datetime, timezone
from enum import Enum
from json import dumps
from typing import Any, Dict, List, Optional, Sequence, Union

from pydantic import BaseModel, ConfigDict, Field, ValidationError, field_validator
from pydantic.types import UUID4
from pydantic_core.core_schema import ValidationInfo

from galileo_core.exceptions.execution import (
    MetricComputingError,
    MetricErrorError,
    MetricFailedError,
    MetricNotApplicableError,
    MetricNotComputedError,
    MetricNotFoundError,
    MetricPendingError,
)
from galileo_core.helpers.logger import logger
from galileo_core.helpers.scorers import SCORER_TO_ALL_METRICS
from galileo_core.schemas.logging.llm import Message
from galileo_core.schemas.shared.document import Document
from galileo_core.schemas.shared.scorers.scorer_name import ScorerName
from galileo_core.utils.json import PydanticJsonEncoder

StepAllowedInputType = Union[str, Sequence[Message]]
StepAllowedOutputType = Union[str, Message, Sequence[Document]]


class StepType(str, Enum):
    llm = "llm"
    retriever = "retriever"
    tool = "tool"
    workflow = "workflow"
    agent = "agent"
    trace = "trace"
    session = "session"


SPAN_TYPES = [StepType.llm, StepType.retriever, StepType.tool, StepType.workflow, StepType.agent]

SPAN_TYPES_WITH_CHILD_SPANS = [StepType.workflow, StepType.agent, StepType.retriever, StepType.tool]

STEP_TYPES_WITH_CHILD_SPANS = [StepType.trace, StepType.workflow, StepType.agent, StepType.retriever, StepType.tool]


class Metrics(BaseModel):
    duration_ns: Optional[int] = Field(
        default=None, description="Duration of the trace or span in nanoseconds.  Displayed as 'Latency' in Galileo."
    )

    model_config = ConfigDict(extra="allow")

    def __getitem__(self, metric: Union[ScorerName, str]) -> Any:
        """
        Get metric value by scorer name or metric name.
        In the case of preset scorers, will convert to the metric name.

        Before returning the value, checks the metric's status. If the status
        indicates a failure condition, raises an appropriate exception.

        Examples:
        metric = Metrics(groundedness=0.85, length1=0.75)
        value = metric[ScorerName.context_adherence]  # returns 0.85
        value = metric["length1"]  # returns 0.75

        Parameters
        ----------
        scorer : Union[ScorerName, str]
            The scorer name or metric name to retrieve.

        Returns
        -------
        Any
            The metric value.

        Raises
        ------
        MetricNotFoundError
            If the metric does not exist in this Metrics object.
        MetricFailedError
            If the metric has 'failed' status.
        MetricErrorError
            If the metric has 'error' status.
        MetricNotComputedError
            If the metric has 'not_computed' status.
        MetricNotApplicableError
            If the metric has 'not_applicable' status.
        MetricPendingError
            If the metric has 'pending' status (still queued).
        MetricComputingError
            If the metric has 'computing' status (still in progress).
        """
        # Resolve the metric name
        if isinstance(metric, ScorerName):
            metric_keys = SCORER_TO_ALL_METRICS.get(metric.value)
            if metric_keys is not None and len(metric_keys) == 1:
                metric_name = metric_keys[0]
            elif metric_keys is not None:
                logger.info(f"{len(metric_keys)} metric keys found for scorer {metric}. Returning None.")
                return None
            else:
                metric_name = metric.value
            user_facing_name = metric.value  # Use scorer name for error messages
        else:
            metric_name = metric
            user_facing_name = metric

        # Check the status before returning the value
        status_key = f"{metric_name}_status"
        status_value = getattr(self, status_key, None)

        if status_value == "failed":
            raise MetricFailedError(user_facing_name)
        elif status_value == "error":
            raise MetricErrorError(user_facing_name)
        elif status_value == "not_computed":
            raise MetricNotComputedError(user_facing_name)
        elif status_value == "not_applicable":
            raise MetricNotApplicableError(user_facing_name)
        elif status_value == "pending":
            raise MetricPendingError(user_facing_name)
        elif status_value == "computing":
            raise MetricComputingError(user_facing_name)

        # Return the metric value
        try:
            return getattr(self, metric_name)
        except AttributeError:
            raise MetricNotFoundError(user_facing_name)


class BaseStep(BaseModel):
    type: StepType = Field(description="Type of the trace, span or session.")
    input: Optional[StepAllowedInputType] = Field(
        default=None, description="Input to the trace or span.", union_mode="left_to_right"
    )
    redacted_input: Optional[StepAllowedInputType] = Field(
        default=None, description="Redacted input of the trace or span.", union_mode="left_to_right"
    )
    output: Optional[StepAllowedOutputType] = Field(
        default=None, description="Output of the trace or span.", union_mode="left_to_right"
    )
    redacted_output: Optional[StepAllowedOutputType] = Field(
        default=None, description="Redacted output of the trace or span.", union_mode="left_to_right"
    )
    name: str = Field(default="", description="Name of the trace, span or session.", validate_default=True)
    created_at: datetime = Field(
        title="Created",
        default_factory=lambda: datetime.now(tz=timezone.utc),
        description="Timestamp of the trace or span's creation.",
    )
    user_metadata: Dict[str, str] = Field(
        default_factory=dict, description="Metadata associated with this trace or span."
    )
    tags: List[str] = Field(default_factory=list, description="Tags associated with this trace or span.")
    status_code: Optional[int] = Field(
        default=None, description="Status code of the trace or span. Used for logging failure or error states."
    )
    metrics: Metrics = Field(default_factory=Metrics, description="Metrics associated with this trace or span.")
    external_id: Optional[str] = Field(default=None, description="A user-provided session, trace or span ID.")
    dataset_input: Optional[str] = Field(
        default=None, title="Dataset Input", description="Input to the dataset associated with this trace"
    )
    dataset_output: Optional[str] = Field(
        default=None, title="Dataset Output", description="Output from the dataset associated with this trace"
    )
    dataset_metadata: Dict[str, str] = Field(
        default_factory=dict,
        title="Dataset Metadata",
        description="Metadata from the dataset associated with this trace",
    )
    id: Optional[UUID4] = Field(title="ID", default=None, description="Galileo ID of the session, trace or span")
    session_id: Optional[UUID4] = Field(
        default=None,
        title="Session ID",
        description="Galileo ID of the session containing the trace or span or session",
    )
    trace_id: Optional[UUID4] = Field(
        default=None,
        title="Trace ID",
        description="Galileo ID of the trace containing the span (or the same value as id for a trace)",
    )

    model_config = ConfigDict(validate_assignment=True)

    def __init__(self, **data: Any):
        for k, v in list(data.items()):
            if v is None:
                del data[k]
        super().__init__(**data)

    @field_validator("name", mode="before")
    def set_name(cls, value: Optional[str], info: ValidationInfo) -> str:
        if value is not None:
            return value
        if "type" in info.data:
            return info.data["type"]
        raise ValidationError("could not set step name from type since type is missing")

    @field_validator("input", mode="after")
    def validate_input_serializable(cls, val: StepAllowedInputType) -> StepAllowedInputType:
        # Make sure we can dump input/output to json string.
        dumps(val, cls=PydanticJsonEncoder)
        return val

    @field_validator("output", mode="after")
    def validate_output_serializable(cls, val: StepAllowedOutputType) -> StepAllowedOutputType:
        # Make sure we can dump input/output to json string.
        dumps(val, cls=PydanticJsonEncoder)
        return val

    @field_validator("user_metadata", mode="before")
    def set_user_metadata(cls, value: Dict[str, Optional[str]]) -> Dict[str, str]:
        return {k: v for k, v in value.items() if v is not None}



from enum import Enum


class AgentType(str, Enum):
    default = "default"
    planner = "planner"
    react = "react"
    reflection = "reflection"
    router = "router"
    classifier = "classifier"
    supervisor = "supervisor"
    judge = "judge"



from typing import Iterable, Set

from galileo_core.schemas.shared.multimodal import ContentModality, MultimodalCapability, MultimodalCapabilityMapping


def supported_capabilities(modalities: Iterable[ContentModality]) -> Set[MultimodalCapability]:
    """
    Returns the set of supported multimodal capabilities based on the provided modalities.
    A multimodal capability is supported only if ALL of its modalities are present.
    """
    modalities_set = set(modalities) if not isinstance(modalities, set) else modalities
    return {
        capability
        for capability, required_modalities in MultimodalCapabilityMapping.items()
        if modalities_set >= required_modalities
    }



from collections import deque
from contextvars import ContextVar
from datetime import datetime
from json import dumps
from typing import Any, Deque, Dict, List, Mapping, Optional

from pydantic import BaseModel, ConfigDict, Field, PrivateAttr
from pydantic.types import UUID4

from galileo_core.schemas.logging.agent import AgentType
from galileo_core.schemas.logging.llm import Event
from galileo_core.schemas.logging.span import (
    AgentSpan,
    LlmMetrics,
    LlmSpan,
    LlmSpanAllowedInputType,
    LlmSpanAllowedOutputType,
    RetrieverSpan,
    Span,
    StepWithChildSpans,
    ToolSpan,
    WorkflowSpan,
)
from galileo_core.schemas.logging.step import Metrics
from galileo_core.schemas.logging.trace import Trace
from galileo_core.schemas.shared.document import Document
from galileo_core.utils.json import PydanticJsonEncoder


class TracesLogger(BaseModel):
    """Logger for managing traces and spans with async-safe parent tracking.

    Each TracesLogger instance maintains its own isolated ContextVar for tracking
    the current parent in the trace hierarchy. This allows:
    - Multiple TracesLogger instances to operate independently
    - Async-safe parent tracking within each instance (via ContextVar semantics)
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    traces: List[Trace] = Field(default_factory=list, description="List of traces.")

    # Per-instance ContextVar for async-safe parent tracking.
    # Each instance has its own ContextVar, providing isolation between logger instances.
    _parent_context_var: ContextVar[Optional[StepWithChildSpans]] = PrivateAttr()

    def _create_context_var(self) -> None:
        """Create a new ContextVar for this instance.

        Each instance must have its own ContextVar to maintain isolation.
        """
        self._parent_context_var = ContextVar(f"_current_parent_{id(self)}", default=None)

    def model_post_init(self, __context: Any) -> None:
        """Initialize the per-instance ContextVar after model construction."""
        self._create_context_var()

    def model_copy(self, *, update: Optional[Mapping[str, Any]] = None, deep: bool = False) -> "TracesLogger":
        """Override model_copy to create a fresh ContextVar on the copy.

        Pydantic's model_copy bypasses model_post_init and would share the
        same ContextVar reference, so we must create a new one for isolation.
        """
        copied = super().model_copy(update=update, deep=deep)
        copied._create_context_var()
        return copied

    def __copy__(self) -> "TracesLogger":
        """Create fresh ContextVar when using copy.copy()."""
        copied = super().__copy__()
        copied._create_context_var()
        return copied

    def __deepcopy__(self, memo: Optional[Dict[int, Any]] = None) -> "TracesLogger":
        """Create fresh ContextVar when using copy.deepcopy().

        ContextVar objects cannot be pickled, so we must handle deepcopy
        by creating a new instance with deepcopied data, then creating
        a fresh ContextVar.
        """
        import copy

        if memo is None:
            memo = {}

        # Deepcopy the traces (the main data)
        copied_traces = copy.deepcopy(self.traces, memo)

        # Create new instance with copied data
        copied = TracesLogger(traces=copied_traces)
        # model_post_init already creates the ContextVar for new instances

        # Register in memo to handle circular references
        memo[id(self)] = copied

        return copied

    def _set_current_parent(self, parent: Optional[StepWithChildSpans]) -> None:
        """Set the current parent for the current async context."""
        self._parent_context_var.set(parent)

    def current_parent(self) -> Optional[StepWithChildSpans]:
        """Get the current parent span/trace for the current async context."""
        return self._parent_context_var.get()

    def reset_parent_tracking(self) -> None:
        """Reset the current parent tracking for this logger instance.

        Call this when reinitializing the logger or switching contexts.
        """
        self._parent_context_var.set(None)

    @property
    def _parent_stack(self) -> Deque[StepWithChildSpans]:
        """Reconstruct the parent stack by walking up the _parent chain.

        This property provides backward compatibility for code that accesses
        _parent_stack directly. The returned deque represents the path from
        root to current position.

        Note: This is a read-only view. Mutations to this deque will not
        affect the actual parent tracking. Use the span methods instead.
        """
        path: List[StepWithChildSpans] = []
        current = self.current_parent()
        while current is not None:
            path.append(current)
            current = current._parent
        # Reverse to get root-to-current order (like a stack bottom-to-top)
        return deque(reversed(path))

    @_parent_stack.setter
    def _parent_stack(self, value: Deque[StepWithChildSpans]) -> None:
        """Set the current parent from a deque (for backward compatibility).

        If deque is empty, clears the current parent.
        Otherwise, sets current parent to the last item and rebuilds _parent chain.
        """
        if not value:
            self._set_current_parent(None)
            return

        # Build the _parent chain from the deque
        items = list(value)
        for i, item in enumerate(items):
            if i == 0:
                item._parent = None
            else:
                item._parent = items[i - 1]

        # Set current parent to the top of the stack (last item)
        self._set_current_parent(items[-1])

    def add_child_span_to_parent(self, span: Span) -> None:
        current_parent = self.current_parent()
        if current_parent is None:
            raise ValueError("A trace needs to be created in order to add a span.")
        span.dataset_input = current_parent.dataset_input
        span.dataset_output = current_parent.dataset_output
        span.dataset_metadata = current_parent.dataset_metadata
        current_parent.add_child_span(span)

    def _get_child_span_timestamp(self) -> datetime:
        """Gets a unique timestamp for a child span from its parent."""
        parent = self.current_parent()
        if not parent:
            raise ValueError("A parent trace or span is required to generate a timestamp.")
        return parent._generate_unique_timestamp()

    def add_trace(
        self,
        input: str,
        redacted_input: Optional[str] = None,
        output: Optional[str] = None,
        redacted_output: Optional[str] = None,
        name: Optional[str] = None,
        created_at: Optional[datetime] = None,
        duration_ns: Optional[int] = None,
        user_metadata: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        dataset_input: Optional[str] = None,
        dataset_output: Optional[str] = None,
        dataset_metadata: Optional[Dict[str, str]] = None,
        external_id: Optional[str] = None,
        id: Optional[UUID4] = None,
    ) -> Trace:
        """
        Create a new trace and add it to the list of traces.
        Simple usage:
        ```
        my_traces.add_trace("input")
        my_traces.add_llm_span("input", "output", model="<my_model>")
        my_traces.conclude("output")
        ```
        Parameters:
        ----------
            input: str: Input to the node.
            redacted_input: Optional[str]: Redacted input to the node.
            output: Optional[str]: Output of the node.
            redacted_output: Optional[str]: Redacted output of the node.
            name: Optional[str]: Name of the trace.
            duration_ns: Optional[int]: Duration of the trace in nanoseconds.
            created_at: Optional[datetime]: Timestamp of the trace's creation.
            user_metadata: Optional[Dict[str, str]]: Metadata associated with this trace.
            ground_truth: Optional[str]: Ground truth, expected output of the trace.
        Returns:
        -------
            Trace: The created trace.
        """
        if self.current_parent() is not None:
            raise ValueError("You must conclude the existing trace before adding a new one.")
        trace = Trace(
            input=input,
            redacted_input=redacted_input,
            output=output,
            redacted_output=redacted_output,
            name=name,
            created_at=created_at,
            user_metadata=user_metadata,
            tags=tags,
            metrics=Metrics(duration_ns=duration_ns),
            dataset_input=dataset_input,
            dataset_output=dataset_output,
            dataset_metadata=dataset_metadata if dataset_metadata is not None else {},
            external_id=external_id,
            id=id,
        )
        trace._parent = None  # Trace is root
        self.traces.append(trace)
        self._set_current_parent(trace)
        return trace

    def add_single_llm_span_trace(
        self,
        input: LlmSpanAllowedInputType,
        output: LlmSpanAllowedOutputType,
        model: Optional[str],
        redacted_input: Optional[LlmSpanAllowedInputType] = None,
        redacted_output: Optional[LlmSpanAllowedOutputType] = None,
        tools: Optional[List[Dict]] = None,
        events: Optional[List[Event]] = None,
        name: Optional[str] = None,
        created_at: Optional[datetime] = None,
        duration_ns: Optional[int] = None,
        user_metadata: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        num_input_tokens: Optional[int] = None,
        num_output_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        status_code: Optional[int] = None,
        time_to_first_token_ns: Optional[int] = None,
        dataset_input: Optional[str] = None,
        dataset_output: Optional[str] = None,
        dataset_metadata: Optional[Dict[str, str]] = None,
        trace_id: Optional[UUID4] = None,
        span_id: Optional[UUID4] = None,
        span_step_number: Optional[int] = None,
    ) -> Trace:
        """
        Create a new trace with a single span and add it to the list of traces.

        Parameters:
        ----------
            input: LlmStepAllowedIOType: Input to the node.
            output: LlmStepAllowedIOType: Output of the node.
            model: str: Model used for this span. Feedback from April: Good docs about what model names we use.
            redacted_input: Optional[LlmStepAllowedIOType]: Redacted input to the node.
            redacted_output: Optional[LlmStepAllowedIOType]: Redacted output of the node
            tools: Optional[List[Dict]]: List of available tools passed to LLM on invocation.
            events: Optional[List[Event]]: List of reasoning, internal tool call, or MCP events.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at: Optional[datetime]: Timestamp of the span's creation.
            user_metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            num_input_tokens: Optional[int]: Number of input tokens.
            num_output_tokens: Optional[int]: Number of output tokens.
            total_tokens: Optional[int]: Total number of tokens.
            temperature: Optional[float]: Temperature used for generation.
            ground_truth: Optional[str]: Ground truth, expected output of the workflow.
            status_code: Optional[int]: Status code of the node execution.
            time_to_first_token_ns: Optional[int]: Time until the first token was returned.
            dataset_input: Optional[str]: Dataset input for the span.
            dataset_output: Optional[str]: Dataset output for the span.
            dataset_metadata: Optional[Dict[str, str]]: Dataset metadata for the span.
            trace_id: Optional[UUID4]: ID of the trace.
            span_id: Optional[UUID4]: ID of the span.
            span_step_number: Optional[int]: Step number of the span.
        Returns:
        -------
            Trace: The created trace.
        """
        if self.current_parent() is not None:
            raise ValueError("A trace cannot be created within a parent trace or span, it must always be the root.")

        trace = Trace(
            input=dumps(input, cls=PydanticJsonEncoder),
            redacted_input=dumps(redacted_input, cls=PydanticJsonEncoder) if redacted_input else None,
            output=dumps(output, cls=PydanticJsonEncoder),
            redacted_output=dumps(redacted_output, cls=PydanticJsonEncoder) if redacted_output else None,
            name=name,
            created_at=created_at,
            user_metadata=user_metadata,
            tags=tags,
            dataset_input=dataset_input,
            dataset_output=dataset_output,
            dataset_metadata=dataset_metadata if dataset_metadata is not None else {},
            id=trace_id,
        )
        trace.add_child_span(
            LlmSpan(
                name=name,
                created_at=created_at,
                user_metadata=user_metadata,
                tags=tags,
                input=input,
                redacted_input=redacted_input,
                output=output,
                redacted_output=redacted_output,
                metrics=LlmMetrics(
                    duration_ns=duration_ns,
                    num_input_tokens=num_input_tokens,
                    num_output_tokens=num_output_tokens,
                    num_total_tokens=total_tokens,
                    time_to_first_token_ns=time_to_first_token_ns,
                ),
                tools=tools,
                events=events,
                model=model,
                temperature=temperature,
                status_code=status_code,
                dataset_input=dataset_input,
                dataset_output=dataset_output,
                dataset_metadata=dataset_metadata if dataset_metadata is not None else {},
                id=span_id,
                step_number=span_step_number,
            )
        )
        self.traces.append(trace)
        # Single span traces are automatically concluded so we reset the current parent.
        self._set_current_parent(None)
        return trace

    def add_llm_span(
        self,
        input: LlmSpanAllowedInputType,
        output: LlmSpanAllowedOutputType,
        model: Optional[str],
        redacted_input: Optional[LlmSpanAllowedInputType] = None,
        redacted_output: Optional[LlmSpanAllowedOutputType] = None,
        tools: Optional[List[Dict]] = None,
        events: Optional[List[Event]] = None,
        name: Optional[str] = None,
        created_at: Optional[datetime] = None,
        duration_ns: Optional[int] = None,
        user_metadata: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        num_input_tokens: Optional[int] = None,
        num_output_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        status_code: Optional[int] = None,
        time_to_first_token_ns: Optional[int] = None,
        id: Optional[UUID4] = None,
        step_number: Optional[int] = None,
    ) -> LlmSpan:
        """
        Add a new llm span to the current parent.

        Parameters:
        ----------
            input: LlmStepAllowedIOType: Input to the node.
            output: LlmStepAllowedIOType: Output of the node.
            model: str: Model used for this span.
            redacted_input: Optional[LlmStepAllowedIOType]: Redacted input to the node.
            redacted_output: Optional[LlmStepAllowedIOType]: Redacted output of the node
            tools: Optional[List[Dict]]: List of available tools passed to LLM on invocation.
            events: Optional[List[Event]]: List of reasoning, internal tool call, or MCP events.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at: Optional[datetime]: Timestamp of the span's creation.
            user_metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            num_input_tokens: Optional[int]: Number of input tokens.
            num_output_tokens: Optional[int]: Number of output tokens.
            total_tokens: Optional[int]: Total number of tokens.
            temperature: Optional[float]: Temperature used for generation.
            status_code: Optional[int]: Status code of the node execution.
            time_to_first_token_ns: Optional[int]: Time until the first token was returned.
            id: Optional[UUID4]: ID of the span.
            step_number: Optional[int]: Step number of the span.
        Returns:
        -------
            LlmSpan: The created span.
        """

        span = LlmSpan(
            input=input,
            redacted_input=redacted_input,
            output=output,
            redacted_output=redacted_output,
            name=name,
            created_at=self._get_child_span_timestamp() if created_at is None else created_at,
            user_metadata=user_metadata,
            tags=tags,
            metrics=LlmMetrics(
                duration_ns=duration_ns,
                num_input_tokens=num_input_tokens,
                num_output_tokens=num_output_tokens,
                num_total_tokens=total_tokens,
                time_to_first_token_ns=time_to_first_token_ns,
            ),
            tools=tools,
            events=events,
            model=model,
            temperature=temperature,
            status_code=status_code,
            id=id,
            step_number=step_number,
        )
        self.add_child_span_to_parent(span)
        return span

    def add_retriever_span(
        self,
        input: str,
        documents: List[Document],
        redacted_input: Optional[str] = None,
        redacted_documents: Optional[List[Document]] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at: Optional[datetime] = None,
        user_metadata: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        status_code: Optional[int] = None,
        id: Optional[UUID4] = None,
        step_number: Optional[int] = None,
    ) -> RetrieverSpan:
        """
        Add a new retriever span to the current parent.

        Parameters:
        ----------
            input: str: Input to the node.
            documents: List[Document]: Documents retrieved from the retriever.
            redacted_input: Optional[str]: Redacted input to the node.
            redacted_documents: Optional[List[Document]]: Redacted documents retrieved from the retriever.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at: Optional[datetime]: Timestamp of the span's creation.
            user_metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            status_code: Optional[int]: Status code of the node execution.
            id: Optional[UUID4]: ID of the span.
            step_number: Optional[int]: Step number of the span.
        Returns:
        -------
            RetrieverSpan: The created span.
        """
        span = RetrieverSpan(
            input=input,
            redacted_input=redacted_input,
            output=documents,
            redacted_output=redacted_documents,
            name=name,
            created_at=self._get_child_span_timestamp() if created_at is None else created_at,
            user_metadata=user_metadata,
            tags=tags,
            status_code=status_code,
            metrics=Metrics(duration_ns=duration_ns),
            id=id,
            step_number=step_number,
        )
        self.add_child_span_to_parent(span)
        return span

    def add_tool_span(
        self,
        input: str,
        redacted_input: Optional[str] = None,
        output: Optional[str] = None,
        redacted_output: Optional[str] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at: Optional[datetime] = None,
        user_metadata: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        status_code: Optional[int] = None,
        tool_call_id: Optional[str] = None,
        id: Optional[UUID4] = None,
        step_number: Optional[int] = None,
    ) -> ToolSpan:
        """
        Add a new tool span to the current parent.

        Parameters:
        ----------
            input: str: Input to the node.
            redacted_input: Optional[str]: Redacted input to the node.
            output: Optional[str]: Output of the node.
            redacted_output: Optional[str]: Redacted output of the node.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at: Optional[datetime]: Timestamp of the span's creation.
            user_metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            status_code: Optional[int]: Status code of the node execution.
            tool_call_id: Optional[str]: ID of the tool call.
            id: Optional[UUID4]: ID of the span.
            step_number: Optional[int]: Step number of the span.
        Returns:
        -------
            ToolSpan: The created span.
        """
        span = ToolSpan(
            input=input,
            redacted_input=redacted_input,
            output=output,
            redacted_output=redacted_output,
            name=name,
            created_at=self._get_child_span_timestamp() if created_at is None else created_at,
            user_metadata=user_metadata,
            tags=tags,
            status_code=status_code,
            tool_call_id=tool_call_id,
            metrics=Metrics(duration_ns=duration_ns),
            id=id,
            step_number=step_number,
        )
        self.add_child_span_to_parent(span)
        return span

    def add_workflow_span(
        self,
        input: str,
        redacted_input: Optional[str] = None,
        output: Optional[str] = None,
        redacted_output: Optional[str] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at: Optional[datetime] = None,
        user_metadata: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        id: Optional[UUID4] = None,
        step_number: Optional[int] = None,
    ) -> WorkflowSpan:
        """
        Add a workflow span to the current parent. This is useful when you want to create a nested workflow span
        within the trace or current workflow span. The next span you add will be a child of the current parent. To
        move out of the nested workflow, use conclude().

        Parameters:
        ----------
            input: str: Input to the node.
            redacted_input: Optional[str]: Redacted input to the node.
            output: Optional[str]: Output of the node. This can also be set on conclude().
            redacted_output: Optional[str]: Redacted output of the node.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at: Optional[datetime]: Timestamp of the span's creation.
            user_metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            id: Optional[UUID4]: ID of the span.
            step_number: Optional[int]: Step number of the span.
        Returns:
        -------
            WorkflowSpan: The created span.
        """
        parent = self.current_parent()
        span = WorkflowSpan(
            input=input,
            redacted_input=redacted_input,
            output=output,
            redacted_output=redacted_output,
            name=name,
            created_at=self._get_child_span_timestamp() if created_at is None else created_at,
            user_metadata=user_metadata,
            tags=tags,
            metrics=Metrics(duration_ns=duration_ns),
            id=id,
            step_number=step_number,
        )
        span._parent = parent  # Set back-pointer before adding to parent
        self.add_child_span_to_parent(span)
        self._set_current_parent(span)
        return span

    def add_agent_span(
        self,
        input: str,
        redacted_input: Optional[str] = None,
        output: Optional[str] = None,
        redacted_output: Optional[str] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at: Optional[datetime] = None,
        user_metadata: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        agent_type: Optional[AgentType] = None,
        id: Optional[UUID4] = None,
        step_number: Optional[int] = None,
    ) -> AgentSpan:
        """
        Parameters:
        ----------
            input: str: Input to the node.
            redacted_input: Optional[str]: Redacted input to the node.
            output: Optional[str]: Output of the node. This can also be set on conclude().
            redacted_output: Optional[str]: Redacted output of the node.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at: Optional[datetime]: Timestamp of the span's creation.
            user_metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            agent_type: Optional[AgentType]: Agent type of the span.
            step_number: Optional[int]: Step number of the span.
        Returns
        -------
            AgentSpan: The created span.
        """
        parent = self.current_parent()
        span = AgentSpan(
            input=input,
            redacted_input=redacted_input,
            output=output,
            redacted_output=redacted_output,
            name=name,
            created_at=self._get_child_span_timestamp() if created_at is None else created_at,
            user_metadata=user_metadata,
            tags=tags,
            metrics=Metrics(duration_ns=duration_ns),
            agent_type=agent_type,
            id=id,
            step_number=step_number,
        )
        span._parent = parent  # Set back-pointer before adding to parent
        self.add_child_span_to_parent(span)
        self._set_current_parent(span)
        return span

    def conclude(
        self,
        output: Optional[str] = None,
        redacted_output: Optional[str] = None,
        duration_ns: Optional[int] = None,
        status_code: Optional[int] = None,
    ) -> Optional[StepWithChildSpans]:
        """
        Conclude the current trace or workflow span by setting the output of the current node. In the case of nested
        workflow spans, this will point the workflow back to the parent of the current workflow span.

        Parameters:
        ----------
            output: Optional[str]: Output of the node.
            redacted_output: Optional[str]: Redacted output of the node.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            Optional[StepWithChildSpans]: The parent of the current workflow. None if no parent exists.
        """
        current_parent = self.current_parent()
        if current_parent is None:
            raise ValueError("No existing workflow to conclude.")

        current_parent.output = output or current_parent.output
        if redacted_output is not None:
            current_parent.redacted_output = redacted_output or current_parent.redacted_output
        current_parent.status_code = status_code
        if duration_ns is not None:
            current_parent.metrics.duration_ns = duration_ns

        # Navigate up to parent via _parent pointer
        parent = current_parent._parent
        self._set_current_parent(parent)

        if parent is None and not isinstance(current_parent, Trace):
            raise ValueError("Finished step is not a trace, but has no parent.  Not added to the list of traces.")
        return self.current_parent()



from operator import attrgetter
from typing import Dict, List, Sequence, cast
from uuid import UUID

from galileo_core.schemas.logging.session import BaseSession, Session
from galileo_core.schemas.logging.span import Span, SpanAdapter
from galileo_core.schemas.logging.step import BaseStep, StepType
from galileo_core.schemas.logging.trace import Trace
from galileo_core.schemas.shared.records import RecordType


def records_to_base_steps(records: Sequence[RecordType]) -> List[BaseStep]:
    """
    Convert RecordType objects to BaseStep objects by validating from dicts.

    This avoids re-validation issues that occur when passing model instances
    directly to create_tree_for_trace/create_tree_for_session. By converting
    to dicts first, Pydantic validates fresh data instead of re-validating
    existing model instances.

    Args:
        records: Sequence of RecordType objects (TraceRecord, SpanRecord, SessionRecord, etc.)

    Returns:
        List of BaseStep objects suitable for passing to create_tree_for_trace/create_tree_for_session
    """
    result: List[BaseStep] = []
    for record in records:
        record_dict = record.model_dump()
        record_type = record_dict.get("type")

        if record_type == StepType.trace:
            # Validate as Trace directly so create_tree_for_trace can use it
            result.append(Trace.model_validate(record_dict))
        elif record_type == StepType.session:
            # Validate as BaseSession - create_tree_for_session will convert to Session
            result.append(BaseSession.model_validate(record_dict))
        else:
            # For spans, validate through SpanAdapter to get correct span type
            result.append(SpanAdapter.validate_python(record_dict))
    return result


def create_tree_for_trace(records: Sequence[BaseStep]) -> Sequence[Trace]:
    parent_id_to_child_ids: Dict[UUID, List[UUID]] = {}
    trace_id_to_trace: Dict[UUID, Trace] = {}
    span_id_to_span: Dict[UUID, Span] = {}
    for record in records:
        if record.id is None:
            raise ValueError("Node ID cannot be None")
        if record.type is StepType.trace:
            trace_id_to_trace[record.id] = Trace.model_validate(record)
        elif record.type is not StepType.session:
            span_id_to_span[record.id] = cast(Span, record)
            if (
                hasattr(record, "parent_id") and record.parent_id is not None
            ):  # Only add value to the parent-child relationship if there is a parent
                if record.parent_id not in parent_id_to_child_ids:
                    parent_id_to_child_ids[record.parent_id] = []
                parent_id_to_child_ids[record.parent_id].append(record.id)

    def populate_child_spans(id: UUID) -> List[Span]:
        child_spans = []
        for child_id in parent_id_to_child_ids.get(id, []):
            span = span_id_to_span[child_id]
            # Only call add_child_spans on spans that support it
            if hasattr(span, "add_child_spans"):
                span.add_child_spans(populate_child_spans(child_id))
            child_spans.append(span)
        return sorted(child_spans, key=lambda record: record.created_at)

    # Build the tree recursively
    traces = []
    for trace_id, trace in trace_id_to_trace.items():
        trace.add_child_spans(populate_child_spans(trace_id))
        traces.append(trace)

    return traces


def create_tree_for_session(records: Sequence[BaseStep]) -> List[Session]:
    session_id_to_session: Dict[UUID, Session] = {}
    session_id_to_children: Dict[UUID, List[BaseStep]] = {}
    for record in records:
        if record.id is None:
            raise ValueError("Node ID cannot be None")
        if record.type is StepType.session:
            session = BaseSession.model_validate(record)
            session_id_to_session[record.id] = Session(**session.model_dump(), traces=[])
        else:
            if (
                hasattr(record, "session_id") and record.session_id is not None
            ):  # Only add value to the Session tree if part of a session
                if record.session_id not in session_id_to_children:
                    session_id_to_children[record.session_id] = []
                session_id_to_children[record.session_id].append(record)

    sessions = []
    for session_id, session in session_id_to_session.items():
        traces = sorted(
            create_tree_for_trace(session_id_to_children.get(session_id, [])),
            key=attrgetter("created_at"),
        )
        session.traces = traces
        sessions.append(session)

    return sessions



from functools import partial
from typing import List, Optional

from pydantic import UUID4

from galileo_core.constants.request_method import RequestMethod
from galileo_core.constants.routes import Routes
from galileo_core.exceptions.http import GalileoHTTPException
from galileo_core.helpers.logger import logger
from galileo_core.helpers.pagination import paginated_request
from galileo_core.schemas.base_config import GalileoConfig
from galileo_core.schemas.core.collaboration_role import CollaboratorRole
from galileo_core.schemas.core.collaborator import UserCollaboratorCreate
from galileo_core.schemas.core.integration.user_integration import (
    UserIntegrationCollaboratorResponse,
)


def list_user_integration_collaborators(
    integration_id: UUID4,
    config: Optional[GalileoConfig] = None,
) -> List[UserIntegrationCollaboratorResponse]:
    config = config or GalileoConfig.get()
    logger.debug("Listing users the integration has been shared with…")
    user_integration_collaborators = paginated_request(
        partial(
            config.api_client.request, RequestMethod.GET, Routes.integration_users.format(integration_id=integration_id)
        ),
        "collaborators",
    )
    users = [UserIntegrationCollaboratorResponse.model_validate(user) for user in user_integration_collaborators]
    logger.debug(f"Listed all users, found {len(users)} users.")
    return users


def share_integration_with_user(
    integration_id: UUID4,
    user_id: UUID4,
    config: Optional[GalileoConfig] = None,
) -> UserIntegrationCollaboratorResponse:
    config = config or GalileoConfig.get()
    logger.debug(f"Sharing integration {integration_id} with user {user_id} with role {CollaboratorRole.viewer}...")
    response_dict = config.api_client.request(
        RequestMethod.POST,
        Routes.integration_users.format(integration_id=integration_id),
        json=[UserCollaboratorCreate(user_id=user_id).model_dump(mode="json")],
    )
    user_shared = [UserIntegrationCollaboratorResponse.model_validate(user) for user in response_dict]
    logger.debug(f"Shared integration {integration_id} with user {user_id} with role {CollaboratorRole.viewer}.")

    try:
        return user_shared[0]
    except IndexError:
        raise GalileoHTTPException(
            message="Galileo API returned empty response.",
            status_code=422,
            response_text=f"Response body {user_shared}",
        )


def update_user_integration_collaborator(
    integration_id: UUID4,
    user_id: UUID4,
    role: CollaboratorRole = CollaboratorRole.viewer,
    config: Optional[GalileoConfig] = None,
) -> UserIntegrationCollaboratorResponse:
    config = config or GalileoConfig.get()
    logger.debug(f"Updating integration {integration_id}: assigning role {role} to user {user_id}…")
    response_dict = config.api_client.request(
        RequestMethod.PATCH,
        Routes.integration_user.format(integration_id=integration_id, user_id=user_id),
        json=UserCollaboratorCreate(user_id=user_id, role=role).model_dump(mode="json"),
    )
    user_integration = UserIntegrationCollaboratorResponse.model_validate(response_dict)
    logger.debug(f"Updated integration {integration_id}: assigned role {role} to user {user_id}…")
    return user_integration


def delete_user_integration_collaborator(
    integration_id: UUID4,
    user_id: UUID4,
    config: Optional[GalileoConfig] = None,
) -> None:
    config = config or GalileoConfig.get()
    logger.debug(f"Removing access for user {user_id} from integration {integration_id}")
    config.api_client.request(
        RequestMethod.DELETE, Routes.integration_user.format(integration_id=integration_id, user_id=user_id)
    )
    logger.debug(f"Access for user {user_id} removed from integration {integration_id}.")



from enum import Enum
from typing import Literal, Union

from pydantic import Field
from typing_extensions import Annotated

from galileo_core.schemas.shared.filtered_collection import MapFilter, StringFilter


class ScorerJobFilterNames(str, Enum):
    node_name = "node_name"
    metadata = "metadata"


class NodeNameFilter(StringFilter):
    """
    Filters on node names in scorer jobs.
    """

    name: Literal[ScorerJobFilterNames.node_name] = ScorerJobFilterNames.node_name


class MetadataFilter(MapFilter):
    """
    Filters on metadata key-value pairs in scorer jobs.
    """

    name: Literal[ScorerJobFilterNames.metadata] = ScorerJobFilterNames.metadata


ScorerJobFilter = Annotated[
    Union[NodeNameFilter, MetadataFilter],
    Field(discriminator="name"),
]



from typing import Generator, List

from pydantic import RootModel

from galileo_core.schemas.shared.message import Message


class Messages(RootModel[List[Message]]):
    def __len__(self) -> int:
        return len(self.root)

    def __iter__(self) -> Generator[Message, None, None]:  # type: ignore[override]
        yield from self.root

    def __getitem__(self, item: int) -> Message:
        return self.root[item]



from typing import Literal

from pydantic import BaseModel

from galileo_core.schemas.core.integration.integration_type import LLMIntegration


class MistralIntegrationCreate(BaseModel):
    name: Literal[LLMIntegration.mistral] = LLMIntegration.mistral
    token: str



from pydantic import UUID4, BaseModel

from galileo_core.schemas.core.collaboration_role import CollaboratorRole


class GroupProjectCollaboratorRequest(BaseModel):
    role: CollaboratorRole
    group_id: UUID4


class GroupProjectCollaboratorResponse(GroupProjectCollaboratorRequest):
    group_name: str



from typing import Literal

from galileo_core.schemas.core.integration.aws import BaseAwsIntegrationCreate
from galileo_core.schemas.core.integration.integration_type import LLMIntegration


class AwsSageMakerIntegrationCreate(BaseAwsIntegrationCreate):
    name: Literal[LLMIntegration.aws_sagemaker] = LLMIntegration.aws_sagemaker



from typing import Literal

from pydantic import BaseModel

from galileo_core.schemas.core.integration.integration_type import LLMIntegration


class AnthropicIntegrationCreate(BaseModel):
    name: Literal[LLMIntegration.anthropic] = LLMIntegration.anthropic
    token: str



from types import SimpleNamespace
from typing import Optional

from pydantic import BaseModel

PaginationDefaults = SimpleNamespace(starting_token=0, limit=100)


class PaginationRequest(BaseModel):
    starting_token: int = PaginationDefaults.starting_token
    limit: int = PaginationDefaults.limit


class PaginationResponse(PaginationRequest):
    paginated: bool = False
    next_starting_token: Optional[int] = None



from datetime import datetime
from functools import partial
from typing import List, Optional

from pydantic import UUID4

from galileo_core.constants.request_method import RequestMethod
from galileo_core.constants.routes import Routes
from galileo_core.helpers.logger import logger
from galileo_core.helpers.pagination import paginated_request
from galileo_core.helpers.user import get_current_user
from galileo_core.schemas.base_config import GalileoConfig
from galileo_core.schemas.core.api_key import ApiKeyResponse, CreateApiKeyRequest, CreateApiKeyResponse
from galileo_core.schemas.core.collaboration_role import CollaboratorRole


def create_api_key(
    description: str,
    expires_at: Optional[datetime] = None,
    project_id: Optional[UUID4] = None,
    project_role: Optional[CollaboratorRole] = None,
    config: Optional[GalileoConfig] = None,
) -> CreateApiKeyResponse:
    config = config or GalileoConfig.get()
    logger.debug(f"Creating API key with description {description}...")
    response_dict = config.api_client.request(
        RequestMethod.POST,
        Routes.create_api_key,
        json=CreateApiKeyRequest(
            description=description, expires_at=expires_at, project_id=project_id, project_role=project_role
        ).model_dump(mode="json"),
    )
    response = CreateApiKeyResponse.model_validate(response_dict)
    logger.debug(f"Created API key {response.id} with description {description}.")
    return response


def list_api_keys(config: Optional[GalileoConfig] = None) -> List[ApiKeyResponse]:
    config = config or GalileoConfig.get()
    user = get_current_user()
    logger.debug(f"Listing API keys for user {user.email}...")
    response_dict = paginated_request(
        partial(config.api_client.request, RequestMethod.GET, Routes.get_api_keys.format(user_id=user.id)), "api_keys"
    )
    api_keys = [ApiKeyResponse.model_validate(api_key) for api_key in response_dict]
    logger.debug(f"Listed {len(api_keys)} API keys.")
    return api_keys


def delete_api_key(api_key_id: UUID4, config: Optional[GalileoConfig] = None) -> None:
    config = config or GalileoConfig.get()
    logger.debug(f"Deleting API key {api_key_id}...")
    config.api_client.request(RequestMethod.DELETE, Routes.delete_api_key.format(api_key_id=api_key_id))
    logger.debug(f"Deleted API key {api_key_id}.")



from typing import List
from uuid import UUID

from pydantic import UUID4

NULL_UUID_LITERAL = "00000000-0000-0000-0000-000000000000"
NULL_UUID_V4_LITERAL = "00000000-0000-4000-8000-000000000000"

NULL_UUID_LITERALS: List[str] = [NULL_UUID_LITERAL, NULL_UUID_V4_LITERAL]

NULL_UUID: UUID = UUID(NULL_UUID_LITERAL)
NULL_UUID_V4: UUID4 = UUID(NULL_UUID_V4_LITERAL)

NULL_UUIDS: List[UUID] = [NULL_UUID, NULL_UUID_V4]



import json
from collections import ChainMap
from csv import DictReader
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Dict, List, Tuple, Union
from warnings import warn

from galileo_core.constants.dataset_format import DatasetFormat

JSONValue = Union[str, int, float, bool, None, Dict[str, "JSONValue"], List["JSONValue"]]

DatasetType = Union[Path, str, Dict[str, List[Any]], List[Dict[str, Any]]]


def process_dataset_value(value: Any) -> JSONValue:
    """Convert Python objects to values that can be serialized to JSON."""

    if isinstance(value, (str, int, float, bool)) or value is None:
        return value

    if isinstance(value, (dict, list)):
        return json.dumps(value, default=str)

    return str(value)


def parse_dataset_dict(dataset: Dict[str, List[Any]]) -> List[Dict[str, JSONValue]]:
    key_rows = [[{key: process_dataset_value(value)} for value in values] for key, values in dataset.items()]
    return [dict(ChainMap(*key_row)) for key_row in zip(*key_rows)]


def parse_dataset_list(dataset: List[Dict[str, Any]]) -> List[Dict[str, JSONValue]]:
    return [dict((k, process_dataset_value(v)) for k, v in row.items()) for row in dataset]


def parse_dataset(dataset: DatasetType) -> Tuple[Path, DatasetFormat]:
    """
    Set the dataset path and format.

    If the dataset is a dictionary or list, it will be stored as a temporary file and
    the path to the temporary file will be returned. If the dataset is a path or string,
    the path will be returned. If the dataset is not a dictionary, list, path, or
    string, a ValueError will be raised.

    If the dataset is a dictionary or list then it will be stored as a JSONL file.

    Parameters
    ----------
    dataset : DatasetType
        Dataset as provided by the user.

    Returns
    -------
    Tuple[Path, DatasetFormat]
        Path to the dataset file and the dataset format.

    Raises
    ------
    ValueError
        If the dataset is not a dictionary, list, path, or string.
    ValueError
        If the dataset file is not a CSV or Feather file.
    FileNotFoundError
        If the dataset file does not exist.
    """
    # Set the dataset path.
    if isinstance(dataset, (dict, list)):
        dataset_rows = parse_dataset_dict(dataset) if isinstance(dataset, dict) else parse_dataset_list(dataset)
        with NamedTemporaryFile(mode="wt", suffix=f".{DatasetFormat.jsonl}", delete=False) as dataset_file:
            for row in dataset_rows:
                dataset_file.write(json.dumps(row, default=str) + "\n")
            dataset_path = Path(dataset_file.name)
    elif isinstance(dataset, Path):
        dataset_path = dataset
    elif isinstance(dataset, str):
        dataset_path = Path(dataset)
    else:
        raise ValueError(f"Invalid dataset type: '{type(dataset)}'.")
    if not dataset_path.exists():
        raise FileNotFoundError(f"Dataset file '{dataset_path}' does not exist.")

    # Set the dataset format.
    if dataset_path.suffix.endswith(DatasetFormat.csv):
        dataset_format = DatasetFormat.csv
        # Check if we have embedding columns.
        reader = DictReader(dataset_path.open("rt", errors="ignore"))
        for column in reader.fieldnames or []:
            if column.endswith("_embedding"):
                warn(
                    f"Column '{column}' appears to be an embedding column, please use "
                    "a feather file as dataset format instead."
                )
    elif dataset_path.suffix.endswith(DatasetFormat.jsonl):
        dataset_format = DatasetFormat.jsonl
    elif dataset_path.suffix.endswith(DatasetFormat.feather):
        dataset_format = DatasetFormat.feather
    else:
        raise ValueError(
            f"Invalid dataset file: '{dataset_path}'. Supported file types are '.csv', '.feather', and '.jsonl'."
        )
    return dataset_path, dataset_format



from functools import partial
from typing import List, Optional

from pydantic import UUID4

from galileo_core.constants.request_method import RequestMethod
from galileo_core.constants.routes import Routes
from galileo_core.helpers.logger import logger
from galileo_core.helpers.pagination import paginated_request
from galileo_core.schemas.base_config import GalileoConfig
from galileo_core.schemas.core.collaboration_role import CollaboratorRole
from galileo_core.schemas.core.user_project import UserProjectCollaboratorRequest, UserProjectCollaboratorResponse


def share_project_with_user(
    project_id: UUID4,
    user_id: UUID4,
    role: CollaboratorRole = CollaboratorRole.viewer,
    config: Optional[GalileoConfig] = None,
) -> UserProjectCollaboratorResponse:
    config = config or GalileoConfig.get()
    logger.debug(f"Sharing project {project_id} with user {user_id} with role {role}...")
    response_dict = config.api_client.request(
        RequestMethod.POST,
        Routes.project_users.format(project_id=project_id),
        json=[UserProjectCollaboratorRequest(user_id=user_id, role=role).model_dump(mode="json")],
    )
    user_shared = [UserProjectCollaboratorResponse.model_validate(user) for user in response_dict]
    logger.debug(f"Shared project {project_id} with user {user_id} with role {role}.")
    return user_shared[0]


def unshare_project_with_user(
    project_id: UUID4,
    user_id: UUID4,
    config: Optional[GalileoConfig] = None,
) -> None:
    config = config or GalileoConfig.get()
    logger.debug(f"Removing access for user {user_id} from project {project_id}")
    config.api_client.request(RequestMethod.DELETE, Routes.project_user.format(project_id=project_id, user_id=user_id))
    logger.debug(f"Access for user {user_id} removed from project {project_id}.")


def list_user_project_collaborators(
    project_id: UUID4, config: Optional[GalileoConfig] = None
) -> List[UserProjectCollaboratorResponse]:
    config = config or GalileoConfig.get()
    logger.debug("Listing users the project has been shared with…")
    user_project_collaborators = paginated_request(
        partial(config.api_client.request, RequestMethod.GET, Routes.project_users.format(project_id=project_id)),
        "collaborators",
    )
    users = [UserProjectCollaboratorResponse.model_validate(user) for user in user_project_collaborators]
    logger.debug(f"Listed all users, found {len(users)} users.")
    return users


def update_user_project_collaborator(
    project_id: UUID4,
    user_id: UUID4,
    role: CollaboratorRole = CollaboratorRole.viewer,
    config: Optional[GalileoConfig] = None,
) -> UserProjectCollaboratorResponse:
    config = config or GalileoConfig.get()
    logger.debug(f"Updating project {project_id}: assigning role {role} to user {user_id}…")
    response_dict = config.api_client.request(
        RequestMethod.PATCH,
        Routes.project_user.format(project_id=project_id, user_id=user_id),
        json=UserProjectCollaboratorRequest(user_id=user_id, role=role).model_dump(mode="json"),
    )
    user_project = UserProjectCollaboratorResponse.model_validate(response_dict)
    logger.debug(f"Updated project {project_id}: assigned role {role} to user {user_id}…")
    return user_project
